(******************************************************************************
*  TParadox Component to Read Paradox 4.0+ .db Files Using File I/O (No BDE)  *
*  vers 1.0                              by Howard Flank flank@speakeasy.net  *
*******************************************************************************

TParadox allows you to read a Paradox 4.0+ table without using the BDE.  It is
short, fast and dumb. It reads the following field types correctly:
    Alpha, Date, Short, Integer, Auto-Increment, Money, Number, Logical, Bytes.
BLOB fields only return the portion stored in the Table (e.g. a "M40" field
would return only the first 40 chars of the Memo).  Time, TimeStamp and BCD
fields are not handled at all.  It can find a record using the Primary Index,
but can't use any secondary indexes. You can install it in the Component Palette
or just put it in your program directory (or a directoruy on the Library path -
 Tools|Environment Options(Library) or Project|Options(Directories)Search Path).
Then use this code snipet (or equivalent) to create the object:
     var Pdx: TParadox;
     ...
     {--Create TParadox Object & Set Properties--}
     Pdx := TParadox.Create(Self);
     Pdx.TableName := 'C:\MyDir\MyTable.db';  Pdx.Name := 'MyPdox';

Properties:
  Active      Table status. Set to True to Open, False to Close.
  BlockSize   Paradox Table BlockSize (after open) r/o (Integer).
  EOF         If attempt to access past last record r/o (Boolean).
  FieldCount  No. of fields in the Table r/o (after open) (Word).
  KeyFields   No. of fields in the Primary Index (after open) (0 if no index).
  RecordCount No. of records in table r/o (after open) (LongWord).
  RecordSize  Paradox record size in bytes (after open) r/o (Word).
  TableName   Filename of the Paradox Table to Read.
Methods:
  Open: Boolean;
    Opens the Table and positions to the first record. True if OK.
  Close;
    Closes the Tables and releases any buffers. Can call if already closed.
  First;
    Positions to the first logical record in the table.
  Next: Boolean;
    Positions to the next logical record in the table. True if OK.
  Field(FldNo or FldName): String;
    Returns the contents of the numbered or named field as text.
  FieldAsInteger(FldNo or FldName): Int64;
    Returns the binary contents of the field as int64 or StrToInt() if String.
  FieldIndex(FldName: String): Integer;
    Returns the field no. of the named field (0=first field, -1=not-found)
  FieldName(FldNo: Integer): String;
    Returns the field name of the number field.
  FindKey([KeyVal1,KeyVal2...]): Boolean;
    Positions to a matching row using the Primary Index (.px file). Time is
    about 20% faster than BDE TTable.FindKey. No. of KeyVals supplied must be
    equal to or less than the no. of fields in the Primary Index. Currently
    FindKey only works for Alpha key fields (or string equivs).
    Returns "True" if a match is Found.
  Locate([FldNo1,FldNo2...],[Value1,Value2...]): Boolean;
    Positions to the Next Table Row that Matches Single or Multi-Fields. Values
    must be the String equiv. of the Field. Uses a slow full table search.
    Returns "True" if a match is Found. Locate starts with the current record,
    so call "Next" first if you want the next match. Or You could just use:
      While not Pdx.EOF and (CompareText(Pdx.Field(1),FindIt)<>0) Do Pdx.Next;
  CompareBytes(const S1,S2; Len: Integer): Integer;  Internal Byte Compare.
    Similar to CompareStr, but 7x faster.

THIS PROGRAM UNIT IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED. INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE ENTIRE RISK AS TO
QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU. SHOULD THE UNIT PROVE
DEFECTIVE, YOU ASSUME THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION. THE AUTHOR IS NOT LIABLE FOR ANY DAMAGE INCLUDING BUT NOT
LIMITED TO: DAMAGE TO HARDWARE, LOSS OF DATA, LOSS OF TIME, LOSS OF HAIR OR
ANTI-SOCIAL BEHAVIOR. YOU MAY USE ANY OR ALL OF THIS PROGRAM CODE FOR WHATEVER
PURPOSES YOU SO DESIRE AND YOU MAY GIVE THE AUTHOR CREDIT OR NOT. *)

unit Paradox;
interface
uses Windows, Messages, SysUtils, Classes, Math;

type {--Types for TParadox routine--}
  APdoxHdr = Packed Record {Paradox Table Header Info}
    RcdLen,HdrLen: Word;  FileType,Blk: Byte;  Rcds: LongWord;
    Used,Total,First,Last: Word;  Fill1: array[$12..$1D] of Byte;
    Root: Word;  Levls: Byte;  Flds,Keys, Curr: Word;  Vers: Byte;   end;
    {FileType: 0=Indexed(.db), 1=PrimaryIndex(.px), 2=NonIndexed(.db) }
    {Vers: 03h=3.0, 04h=3.5, 05h-09h=4.0, 0ah-0Bh=5.0, 0Ch=7.0}
  PPdoxBlk = ^APdoxBlk;
  APdoxBlk = Packed Record {Paradox Data Block Header}
    Next,Prev,Last: Word;   end;
  APdoxBlob = Packed Record {10-byte Blob Info Block}
    Offset,Length: LongWord;  ModNum: Word;   end;
  APdoxFld = Record
    Name: String[25];  Typ,Start,Len: Integer;   end;
  EParadoxError = Class(Exception);

type TParadox = class(TComponent)
  private
    FTableName: string;
    FBlkSize,IndxSize,RcdIdx: Integer;
    FActive,FEOF: Boolean;
    Pdox: File;  Bfr: string;  PH: APdoxHdr;  DB: APdoxBlk;  BI: APdoxBlob;
    Indx: File;  Idx: string;  PI: APdoxHdr;  PX: APdoxBlk;
    Flds: Array of APdoxFld;
    procedure SetActive(const Value: Boolean);
  protected
    function  ReadDB(BlkNo: Word): Boolean;
    function  ReadPX(BlkNo: Word): Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function  Open: Boolean;
    procedure Close;
    function  Next: Boolean;
    Procedure First;
    function  Field(FldNo: Integer): String; overload;
    function  Field(FldName: String): String; overload;
    function  FieldAsInteger(FldNo: Integer): Int64; overload;
    function  FieldAsInteger(FldName: String): Int64; overload;
    function  FieldIndex(FldName: String): Integer;
    function  FieldName(FldNo: Integer): String;
    function  FindKey(Val: Array of String): Boolean;
    function  Locate(FldNo: Array of Integer;  Val: Array of String): Boolean;
    property  BlockSize: Integer read FBlkSize;
    property  EOF: Boolean read FEOF;
    property  FieldCount: Word read PH.Flds;
    property  KeyFields: Word read PH.Keys;
    property  RecordCount: LongWord read PH.Rcds;
    property  RecordSize: Word read PH.RcdLen;
  published
    property  Active: Boolean read FActive write SetActive;
    property  TableName: string read FTableName write FTableName;
  end;
function CompareBytes(const V1,V2; Len: Integer): Integer;
procedure Register;
implementation

{---- TParadox Constructor and Destructor ------------------------------------}
constructor TParadox.Create(AOwner: TComponent); begin
  inherited;  FActive := False;  FEOF := True;  Flds := nil;
  end{constructor};
destructor TParadox.Destroy;  begin
  Close;  inherited;   end{destructor};

{---- Change the "Active" Status (Open or Close) -----------------------------}
procedure TParadox.SetActive(const Value: Boolean);  begin
  If Value=FActive Then Exit; {No Change}
  If FActive Then Close Else Open;   end{SetActive};

{---- Open the Paradox Table and Get the Field Info --------------------------}
function TParadox.Open: Boolean;
  var i,fs: Integer;  p: PChar;  fm: Byte;
begin   Result := True;
  If TFileRec(Pdox).Mode>fmClosed Then Exit; {Aready Open}
  AssignFile(Pdox,FTableName);  fm := FileMode;  FileMode := fmOpenRead;
  Try  Reset(Pdox,1);  FileMode := fm;  BlockRead(Pdox,PH,SizeOf(PH));
  Except on E:EInOutError Do
    Raise EParadoxError.CreateFmt('%s: Unable to open database "%s" - %s',
        [Self.Name,FTableName,E.Message]);  end{Try};
  FBlkSize := PH.Blk*1024;  SetLength(Bfr,FBlkSize);
  BlockRead(Pdox,Bfr[SizeOf(PH)],PH.HdrLen-SizeOf(PH)-1);
  PH.Vers := Byte(Bfr[$39]);  PH.Curr := 0;
  SetLength(Flds,PH.Flds);  p := @Bfr[6*PH.Flds+203];  fs := 0;
  For i := 0 to PH.Flds-1 Do begin {-Get Fld Name,Type,Len,Start-}
    Flds[i].Name := StrPas(p);  p := p+StrLen(p)+1;  Flds[i].Start := fs;
    Flds[i].Typ := Ord(Bfr[2*i+120]);  Flds[i].Len := Ord(Bfr[2*i+121]);
    If Flds[i].Typ=$17 Then Inc(fs,17)  Else Inc(fs,Flds[i].Len);   end{For i};
  First();  FActive := True;   end{Open};

{---- Close the Database & Index Files and Release Buffers -------------------}
procedure TParadox.Close;  begin
  If TFileRec(Pdox).Mode>fmClosed Then CloseFile(Pdox);
  If TFileRec(Indx).Mode>fmClosed Then CloseFile(Indx);
  SetLength(Bfr,0);  SetLength(Idx,0);  Flds := nil;  FActive := False;
  end{Close};

{---- Position to the First Logical Record in the Table ----------------------}
Procedure TParadox.First;  begin
  FEOF := False;  ReadDB(PH.First);   end{First};

{---- Position to the Next Logical Record in the Table ----------------------}
function TParadox.Next: Boolean;
begin  Result := False;
  If FEOF or not FActive Then Exit;
  If RcdIdx>DB.Last+6 Then {Read Next Block}
    If DB.Next<>0 Then ReadDB(DB.Next)  Else FEOF := True
  Else Inc(RcdIdx,PH.RcdLen); {Next Rcd in Bfr}
  Result := not FEOF;   end{Next};

{---- Get the Contents a Field as Text by Number (0 is the first field) ------}
function TParadox.Field(FldNo: Integer): String;
  var  i,fs,fl: Integer;  pc: PChar;  dt: TDateTime;  s: String[7];
    si: SmallInt absolute s;  int: Integer absolute s;  d: Double absolute s;
begin
  If FEOF or not FActive or (FldNo<0) or (FldNo>=PH.Flds) Then begin
    Result := '<error>';  Exit;  end;
  fs := RcdIdx+Flds[FldNo].Start;  fl := Flds[FldNo].Len;  pc := @Bfr[fs];
  If Flds[FldNo].Typ in [2..6,$14..$16] Then begin {-Cvt Numeric Types-}
    For i := 0 to fl-1 Do  s[fl-i-1] := Bfr[fs+i]; {Invert Low/High Bytes}
    s[fl-1] := Char(Ord(s[fl-1]) xor $80);   end; {invert sign bit}
  Case Flds[FldNo].Typ of
    1: SetString(Result,pc, Min(StrLen(@Bfr[fs]),fl));                  {Alpha}
    2: begin  dt := int-693594;  Result := DateTimeToStr(dt);  end;      {Date}
    3: Result := IntToStr(si);                                          {Short}
    4,$16: Result := IntToStr(int);                             {Integer,+Auto}
    5,6: Result := FloatToStr(d);                                {Money,Number}
    9: If Bfr[fs]=#$80 Then Result := 'False'  Else Result := 'True'; {Logical}
    $14: Result := '<Time>'; {L=4}                                      {*Time}
    $15: Result := '<TimeStamp>'; {L=8}                            {*TimeStamp}
    $17: SetString(Result,pc,17); {Returns the Raw Value}                {*BCD}
    $18: SetString(Result,pc,fl);                                       {Bytes}
    $C..$10: begin  Move(pc[fl-10],BI,10);                              {BLOBs}
      SetString(Result,pc,Min(fl-10,BI.Length));   end;
    Else Result := '<Unknown Type>';
    end{Case};   end{Field};
{---- Get the Contents a Field as Text by FieldName --------------------------}
function TParadox.Field(FldName: String): String;  begin
  Result := Field(FieldIndex(FldName));   end{Field};

{---- Get the Contents a Field as Integer by Field Number --------------------}
function  TParadox.FieldAsInteger(FldNo: Integer): Int64;
  var  i,fs,fl: Integer;  s: String[7];
    si: SmallInt absolute s;  int: Integer absolute s;  d: Double absolute s;
begin
  If FEOF or not FActive or (FldNo<0) or (FldNo>=PH.Flds) Then Result := 0
  Else If Flds[FldNo].Typ in [2..9,$14..$16] Then begin
    fs := RcdIdx+Flds[FldNo].Start;  fl := Flds[FldNo].Len;
    For i := 0 to fl-1 Do  s[fl-i-1] := Bfr[fs+i];
    s[fl-1] := Char(Ord(s[fl-1]) xor $80); {invert sign bit}
    If Flds[FldNo].Typ=2 Then Dec(int,693594); {Cvt date to std}
    If fl=1 Then Result := Byte(si)
    Else If fl=2 Then Result := si
    Else If fl=4 then Result := int
    Else Try Result := Round(d);
      Except On EInvalidOp Do Result := 0;   end{Try};   end{If [numeric]}
  Else Result := StrToIntDef(Field(FldNo),0);   end{FieldAsInteger};
{---- Get the Contents a Field as Integer by FieldName -----------------------}
function  TParadox.FieldAsInteger(FldName: String): Int64;   begin
  Result := FieldAsInteger(FieldIndex(FldName));   end{FieldAsInteger};

{---- Find the Next Table Row that Matches a Multi-Column Key ----------------}
{     Uses a really slow full table search.  Returns "True" if Found.         }
{     Search starts with current record, so call "Next" first for next match. }
function TParadox.Locate(FldNo: Array of Integer;
         Val: Array of String): Boolean;
  var  i: Integer;  Found: Boolean;
begin   Result := False;  Found := False;
  If not FActive or (High(FldNo)<>High(Val)) then Exit;
  While not FEOF and not Found do begin   Found := True;
    For i := 0 to High(Val) Do
      If Field(FldNo[i])<>Val[i] Then begin  Found := False;  Break;  end;
    If not Found Then Next();   end{While};
  Result := Found;   end{Locate};

{---- Search for a Specified Record Using the Primary Key --------------------}
function  TParadox.FindKey(Val: Array of String): Boolean;
  var  b,b1,b2,bn,block,cmp,i,ip,len: Integer;  LookFor: String;  fm: byte;
begin   Result := False;  LookFor := '';
  If not FActive or (PH.Keys=0) or (High(Val)>PH.Keys) then Exit;
  If TFileRec(Indx).Mode<=fmClosed Then begin {-Open the .px file-}
    AssignFile(Indx,Copy(FTableName,1,Pos('.',FTableName))+'px');
    fm := FileMode;  FileMode := fmOpenRead;
    Try  Reset(Indx,1);  FileMode := fm;  BlockRead(Indx,PI,SizeOf(PI));
    Except on E:EInOutError Do
      Raise EParadoxError.CreateFmt('%s: Unable to open primary index %s - %s',
          [Self.Name,TFileRec(Indx).Name,E.Message]);  end{Try};
    IndxSize := PI.Blk*1024;  SetLength(Idx,IndxSize+4);
    BlockRead(Indx,Idx[SizeOf(PI)],PI.HdrLen-SizeOf(PI)-1);
    PI.Curr := 0;   end{If <=fmClosed};
  For i := 0 to High(Val) Do {-Create Key to Match in Index-}
    LookFor := LookFor+Val[i]+StringOfChar(#0,Flds[i].Len-Length(Val[i]));
  len := Length(LookFor);  block := 1;
  For i := 1 to PI.Levls Do begin {-Search down the index levels-}
    ReadPX(Block);  Idx[PX.Last+PI.RcdLen+7] := #$FF;
    bn := PX.Last div PI.RcdLen +1;  b1 := 0;  b2 := bn;
    Repeat {-Binary Search for entry1<LookFor<entry2 in Index-}
      b := (b1+b2) div 2;  ip := b * PI.RcdLen + 7;
      cmp := CompareBytes(LookFor[1],Idx[ip],len);
      If cmp<0 Then b2 := b  Else b1 := b;  Until (b2-b1)<=1;
    If (b1=0) and (CompareBytes(LookFor[1],Idx[7],len)<0) Then Exit; {Too Low}
    ip := b2*PI.RcdLen+7;
    Block := (Ord(Idx[ip-6]) xor $80) shl 8 + Ord(Idx[ip-5]);   end{For i};
  ReadDB(Block);  Dec(RcdIdx,PH.RcdLen);
  Repeat   Inc(RcdIdx,PH.RcdLen); {-Find Key in Rcd-}
    cmp := CompareBytes(LookFor[1],Bfr[RcdIdx],len);
    If cmp=0 Then begin  Result := True;  Break;  end;
    Until (cmp<0) or (RcdIdx>DB.Last+6);
  end{FindKey};

{---- Get the Field Number of a Field by FieldName ---------------------------}
function TParadox.FieldIndex(FldName: String): Integer;
  var i: Integer;  begin
  For i := 0 to PH.Flds-1 Do
    If CompareText(FldName,Flds[i].Name)=0 Then begin
      Result := i;  Exit;   end;
  Result := -1;   end{FieldIndex};

{---- Get the FieldName of a Field by Number ---------------------------------}
function TParadox.FieldName(FldNo: Integer): String;  begin
  If (FldNo>=0) and (FldNo<PH.Flds) Then Result := Flds[FldNo].Name
  Else Result := '<error>';   end{FieldName};

{---- Internal Routines to Read a .db and .px Data Block ---------------------}
function TParadox.ReadDB(BlkNo: Word): Boolean;  begin
  If (BlkNo<1) or (BlkNo>PH.Total) Then begin  Result := False;  Exit;  end;
  If BlkNo<>PH.Curr Then begin
    Seek(Pdox,(BlkNo-1)*FBlkSize+PH.HdrLen);  BlockRead(Pdox,Bfr[1],FBlkSize);
    Move(Bfr[1],DB,6);  PH.Curr := BlkNo;   end;
  RcdIdx := 7;  Result := True;   end{ReadDB};
function TParadox.ReadPX(BlkNo: Word): Boolean;  begin
  If (BlkNo<1) or (BlkNo>PI.Total) Then begin  Result := False;  Exit;  end;
  If BlkNo<>PI.Curr Then begin
    Seek(Indx,(BlkNo-1)*IndxSize+PI.HdrLen);  BlockRead(Indx,Idx[1],IndxSize);
    Move(Idx[1],PX,6);  PI.Curr := BlkNo;   end;
  Result := True;   end{ReadPX};

{-----------------------------------------------------------------------------}
{  COMPAREBYTES Compares <Len> Bytes at V1 with Bytes at V2, Including 00h's  }
{  Returns: 1 if V1>V2, 0 if V1=V2 and -1 if V1<V2                            }
{-----------------------------------------------------------------------------}
function CompareBytes(const V1,V2; Len: Integer): Integer;
  asm		Push	ESI		{Save ESI,EDIX}
		Push	EDI
		Mov	ESI,EAX		{Get Addr of V1}
		Mov	EDI,EDX		{Get Addr of V2, Len already in CX}
		RepE	CmpSB		{Compare v1 and V2}
                Mov     EAX,1		{Set V1>V@ code}
                JA      @Exit
                MOV	EAX,0		{Set V1=V2 code}
                JE	@Exit
                DEC	EAX		{Set V1<V2 code}
  @Exit:	Pop	EDI		{Restore EDI,ESI}
		Pop	ESI
  end{CompareBytes}; {Note: 7x faster than CompareStr}


{---- Register this Weasel with Delphi IDE -----------------------------------}
procedure Register;  begin
  RegisterComponents('Data Access', [TParadox]);   end;

end.
